#line 1 "../main.ino"
#include <Arduino.h>
#line 1
void setup ();
void loop ();
void state_logic (void );
void output_logic (void );
void interrupt_service_routine_INT0 (void );
void read_inputs (void );
void write_outputs (void );
void update_time_flag (void );
#line 1
	/*	===========================================================================
 UPC - EETAC - CSD - http://digsys.upc.edu

 * LAB5: FSM with interrupts
 * Documentation location: https://digsys.upc.edu/csd/DEE/lab05/lab05.html
 * Hardware for prototyping: Arduino Uno board. Chip ATmega328P

Planning: 

- Step 1: Prepare the hardware, so that adapt the Arduino board to this 
particular application. For instance: add two switches and three LED circuits

- Step 2: Prepare the software. Hardware/software diagram. State diagram

    - Real-time measurement using millis() 
=============================================================================*/

/*	===========================================================================
	Include section
	=========================================================================*/

// External function libraries 


/*	===========================================================================
	Definitions
	======================================================================== */
// Switches 
const char pin_SW2 = 5;
const char pin_SW1 = 4;


// Start-stop push-button. Pin PD2 is also the microcontroller's external INT0
const char INT0_pin_ST_BUTTON = 2;  // 

// LEDs for indicating the current state
const char pin_LED2 = 9; 
const char pin_LED1 = 8; 

// LED for generating the squared waveform 
const char pin_W_L = 12; 

 
// FSM list of states
const char IDLE               = 'A'; 
const char One_click_detected = 'B'; 
const char STATE_0            = 'C'; 
const char STATE_1            = 'D'; 
const char STATE_2            = 'E'; 
const char STATE_3            = 'F'; 


// Timer period for generating a 12 Hz intermittency: 
// Waveform will be toggled after every TIMER_PERIOD
const unsigned long TIMER_PERIOD = 41; // 41.6666 ms   (about 24 Hz)


/*	===========================================================================
	Global RAM variables
	======================================================================== */

char   	var_SW1; 
char   	var_SW2; 
					
char   	var_S; // For representing the state in binary
char   	var_Wave = 0; // Variables will be initialised when running setup()

char 	var_TW_flag; // For timing half wave periods

char 	var_current_state ;    // This is the key varible: FSM state memory


char 	var_ST_flag;    // It will be set when one click detected on interrupt

char 	var_USART_flag; // For writing or sending ASCII messages only one time					

// For timing the square waveform: 
static unsigned long var_previous_time;   
 
/*	===========================================================================
	System initialisation
	=========================================================================*/
void setup() { 
// Configuring digital output pins
	pinMode(pin_LED1, OUTPUT);  
	pinMode(pin_LED2, OUTPUT);	
	pinMode(pin_W_L, OUTPUT);	

// Configuring digital input pins
	pinMode(pin_SW1, INPUT);  
	pinMode(pin_SW2, INPUT);  

// This is the special pit to be an external interrupt source INT0:
	pinMode(INT0_pin_ST_BUTTON, INPUT_PULLUP); 
	
// Serial monitor initialisation
	Serial.begin(9600); 
 
 //Initialise the FSM and the timer 
	var_current_state = IDLE;	
	var_previous_time = 0; 	 

		// Serial monitor ASCII message:
	Serial.println("Project LAB5 - DEE");
	Serial.println("FSM + external interrupts"); 
	Serial.println("System IDLE: waiting for two ST clicks to start"); 


//	---------------------------------------------------------------------------
// This is the interrupt handler. Read the Arduino documentation at: 
// https://www.arduino.cc/reference/en/language/functions/external-interrupts/attachinterrupt/
// attachInterrupt(digitalPinToInterrupt(pin), ISR, mode) (recommended)
	
	attachInterrupt(digitalPinToInterrupt(INT0_pin_ST_BUTTON), interrupt_service_routine_INT0, FALLING);
	// Interrupt when detecting a falling edge from '1' to '0'
}

/*	===========================================================================
	Infinite loop 
	======================================================================== */
void loop() {       

	state_logic();    // calculating the next state to go
	output_logic();		// calculating outputs 

}

/*	===========================================================================
	State logic to determine where to go next
	======================================================================== */
void state_logic(void){
	read_inputs(); 		// capture the switch voltage values into RAM variables
	switch (var_current_state) {

	case IDLE:	// Do nothing, wait for the first ST click
		if (var_ST_flag == 1) {	  // jump to the next state
			  var_current_state = One_click_detected;
			  var_ST_flag = 0;	//Clear flag when used
			  var_USART_flag = 1; //For writing a message only one time
		}
		  else
			var_current_state = IDLE;  // loop doing nothing   
		break;

	case One_click_detected:	// Wait for the second ST click
		if (var_ST_flag == 1) {	  // start the process
			  var_current_state = STATE_0;
			  var_ST_flag = 0;	  //Clear flag when used
						// For printing a message after jumping to STATE_0
			  var_USART_flag = 1; 
		}
		else 
			var_current_state = One_click_detected;  // loop   
		break;

	case STATE_0:	
			var_current_state = STATE_1; // jump immediately      
		break;
	
	case STATE_1:	// jump to the next state depending on input reading 
			if (var_SW1 == 1)	var_current_state = STATE_1;   
			else var_current_state = STATE_2; 
		break;

	case STATE_2:	
			var_current_state = STATE_3;     
		break;

	case STATE_3:	
		if (var_ST_flag == 1) {	  // Stop the FSM going to IDLE
			  var_current_state = IDLE;
			  var_ST_flag = 0;	  //Clear flag when used
				var_USART_flag = 1; // Print a message when arriving at IDLE
		}
		else if (var_SW2 == 1) var_current_state = STATE_3;
		else var_current_state = STATE_0; //continue running the process

		break;
	}	
}

/*	===========================================================================
	Output logic 
	======================================================================== */
void output_logic(void){

	switch (var_current_state) {

	case IDLE:
		var_S = 0;
		var_Wave = 0;
		if (var_USART_flag == 1){
			var_USART_flag = 0;
			Serial.println("Idle state ... waiting for ST click to start"); 
		}
		break;

	case One_click_detected:
		var_S = 0;
		var_Wave = 0;
		if (var_USART_flag == 1){
			var_USART_flag = 0;
			Serial.println("One button click detected... waiting a second click"); 
		}
		break;

	case STATE_0:
		var_S = 0;
		var_Wave = 0;

		if (var_USART_flag == 1){
			var_USART_flag = 0;
			Serial.println("FSM operating .... "); 
		}
		break;

	case STATE_1: // Here the square wave has to be generated
		update_time_flag();    //  to keep track of real time
		var_S = 1;
		if (var_TW_flag == 1){
			var_Wave = not(var_Wave); 	// Toggle waveform
			var_TW_flag = 0; 			// clear flag when used
		}
		break;

	case STATE_2:
		var_S = 2;
		var_Wave = 0; 	// Here waveform output is ON
		break;		
	
	case STATE_3:
		var_S = 3;
		var_Wave = 1;
		break;
	}		
	write_outputs(); // Convert RAM variable values into pin output voltage	
}


/*	===========================================================================
	ISR external interrupt INT0 pin 
This function is executed when an active edge is detected at pin PD2 (INT0)
The idea is simply set the global variable
	======================================================================== */
void interrupt_service_routine_INT0 (void){

	var_ST_flag = 1;  

	// LAB experimetation on debouncing techniques: 
	//delay(5); // wait some time 

}


/*	===========================================================================
	read_inputs 
	======================================================================== */
void read_inputs(void){

   if (digitalRead(pin_SW1) == 1) // Active-low connected switch
		var_SW1 = 1;
   else
		var_SW1 = 0;
   
   if (digitalRead(pin_SW2) == 1) // Active-high connected switch
		var_SW2 = 1;
   else
		var_SW2 = 0;
}

/*	===========================================================================
	write_outputs 
	======================================================================== */
void write_outputs(void){
static char var_buf;  // local variable for saving partial results

	var_buf = var_S & 0b00000001; //Masking the bit of interest
   if (var_buf == 1)
       digitalWrite(pin_LED1, HIGH); // Active-high connected LED for S1
   else
       digitalWrite(pin_LED1, LOW);  
	
	var_buf = (var_S & 0b00000010)>>1;
   if (var_buf == 1)
       digitalWrite(pin_LED2, HIGH); // Active-high connected LED for S0
   else
       digitalWrite(pin_LED2, LOW);  
	

   if (var_Wave == 1)
       digitalWrite(pin_W_L, LOW); // Active-low connected LED for W_L
   else
       digitalWrite(pin_W_L, HIGH);  

}


/*	===========================================================================
	Update time to calculate half periods for the output waveform
	======================================================================== */
void update_time_flag(void){
static unsigned long var_current_time;    // current time in ms
static unsigned long var_time_elapsed;   //  time elapsed since last flag
   
    var_current_time = millis();	// From internal timer that counts time
    var_time_elapsed =  var_current_time - var_previous_time;  

	if(var_time_elapsed  > TIMER_PERIOD) {
			// is the flag is set is because time elapsed is sampling period
		var_TW_flag = 1;   
			// to start another sampling period cycle
		var_previous_time = var_current_time; 
	}
}




